### Importing needed libraries ###
import es, cPickle, popuplib
from os.path import lexists
from random import randint
from playerlib import getPlayer
### Setting some globals ###
info = es.AddonInfo()
info.name = "Hangman"
info.version = "1.0.0 Beta"
info.url = "http://addons.eventscripts.com/addons/view/hangman"
info.author = "Ojii"
info.basename = "hangman"
info.description = "Let's play hangman!"
letters = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z']
user = {}
pending = {}
gfx = {}
es.ServerVar('hangman_version', info['version'], 'Hangman Addon by Ojii').makepublic()
ENABLE_ERRORLIB = True
if ENABLE_ERRORLIB:
    import errorlib

### Function called when loading the addon. Game specific settings are set. ###
def load():
    global lines
    es.msg('Loading hangman...')
    if ENABLE_ERRORLIB:
        errorlib.registerAddon(info,'me@ojii.ch')
    if es.getgame() in ['Counter-Strike: Source','Day of Defeat: Source']:
        lines = 8
    else:
        lines = 5
    load_gfx()
    load_db()
    load_pending()
    load_configs()
    load_words()
    es.addons.registerSayFilter(hangman_filter)
    
### On unloading the addon the databases are saved ###
def unload():
    if ENABLE_ERRORLIB:
        errorlib.unregisterAddon(info)
    save_db()
    save_pending()
    es.addons.unregisterSayFilter(hangman_filter)
    es.dbgmsg('Unloaded hangman...')    

### If a player joins the server, we tell him that there is a nice addon on this server and how to use it ###
def player_activate(ev):
    global user
    s = ev['es_steamid']
    u = ev['userid']
    n = ev['es_username']
    if not es.isbot(u):
        if not user.has_key(s):
            user[s] = {'name': n, 'words': [],'won':0,'lost':0, 'enable':1}
            es.tell(u,'Welcome ' + n + '!')
            es.tell(u,'This server uses Hangman ' + info.version + '.')
            es.tell(u,'When you die, you can play Hangman, if you want to disable this type !togglehangman')
            es.tell(u,'Type !hangman for more informations about this addon')
        else:
            r = get_rank(s)
            user[s]['name'] = n
            es.tell(u,'Welcome back ' + n + '!')
            es.tell(u,'Your Hangman statistics:')
            es.tell(u,'Your currently ranked ' + number(r) + ' with ' + str(user[s]['won']) + ' games won and ' + str(user[s]['lost']) + 'games lost.')
            es.tell(u,'Type !hangman for more informations about this addon')

### If a player dies and he has hangman enabled, lets play a game ###
def player_death(ev):
    s = ev['es_steamid']
    u = ev['userid']
    if not es.isbot(u):
        if user[s]['enable'] == 1:
            play(s,u)
        else:
            es.tell(u,'You have currently disabled hangman, to enable it type !togglehangman or type !hangman for more information')

### If a player spawns, interrupt his current game ###
def player_spawn(ev):
    if popuplib.isqueued('hangman_popup', ev['userid']):
        hmp.unsend(ev['userid'])    

### This is the main game function ###
def play(s,u):
    global pending
    if not pending.has_key(s):
        es.tell(u, '#green','Shall we play a game?')
        word = get_word(s)
        wordletters = []
        for x in range(len(word)):
            if not word[x] == '\n':
                wordletters.append(word[x].upper())
        pending[s] = {'word':word,'wordletters':wordletters,'tested':[],'errors':0}
        user[s]['words'].append(word)
    else:
        es.tell(u, '#green','Lets go on!')
    print_word(s,u)

### Using filter over saycommands. This does a lot... ###
def hangman_filter(userid, message, teamonly):
    text = message.strip('"')
    s = es.getplayersteamid(userid)
    if getPlayer(userid).get('isdead') == 1:
        if not pending.has_key(s):
            return (userid,text,teamonly)
        if len(text) != 1:
            return (userid,text,teamonly)
        if not text.upper() in letters:
            return (userid,text,teamonly)
        if not text.upper() in pending[s]['tested']:
            if popuplib.isqueued('hangman_popup', userid):
                hmp.unsend(userid)
            pending[s]['tested'].append(text.upper())
            if text.upper() in pending[s]['wordletters']:
                es.tell(userid, '#green','This letter is in the word!')
                print_word(s,userid)
                if won(s):
                    es.tell(userid, '#green','*** CONGRATULATIONS YOU WON ***')
                    del pending[s]
                    user[s]['won'] = user[s]['won'] + 1
                    play(s,userid)
            else:
                es.tell(userid, '#green','This letter is not in the word!')
                pending[s]['errors'] = pending[s]['errors'] + 1
                if pending[s]['errors'] == 8:
                    es.tell(userid, '#green','### YOU LOST ###')
                    es.tell(userid, '#green','The word was: ' + pending[s]['word'])
                    del pending[s]
                    user[s]['lost'] = user[s]['lost'] + 1
                    play(s,userid)
                else:
                    if lines == 5:
                        fivelines(userid,pending[s]['errors'])
                    else:
                        eightlines(userid,pending[s]['errors'])
            return (0, None, None)
        else:
            es.tell(userid, '#green','You have already tried this letter!')
            return (0, None, None)
    elif text == '!togglehangman':
        if user[s]['enable'] == 1:
            user[s]['enable'] = 0
            es.tell(userid, '#green','Hangman disabled...')
        else:
            user[s]['enable'] = 1
            es.tell(userid, '#green','Hangman enabled...')
        return (0,None,None)
    elif text == '!highscores':
        highscores = get_highscores()
        print_highscores(highscores,s,userid)
        return (0,None,None)
    elif text == '!myrank':
        r = get_rank(s)
        es.tell(userid,'#green','You are ranked ' + number(r) + ' of ' + str(len(user)) + ' players with a score of ' + str(user[s]['won'] / user[s]['lost']) + ' points (' + str(user[s]['won']) + '/' + str(user[s]['lost']) + ')')
        return (0,None,None)
    elif text == '!hangman':
        es.tell(userid,'#green','Hangman commands: !togglehangman (enables or disables hangman), !myrank and !highscores')
        es.tell(userid,'#green','You can play hangman while you are dead. Just type the letter to test it...')
        return (0,None,None)
    else:
        return (userid,text,teamonly)

### Test if a player has won ###
def won(s):
    for letter in pending[s]['wordletters']:
        if not letter in pending[s]['tested']:
            return False
    return True    

### Print the 'word' with the letters already guessed right ###
def print_word(s,u):
    draw = ''
    for x in range(len(pending[s]['wordletters'])):
        if pending[s]['wordletters'][x] in pending[s]['tested']:
            draw = draw + ' ' + pending[s]['wordletters'][x] + ' '
        else:
            draw = draw + ' _ '
    es.tell(u, '#green', draw)

### Output the highscore ###
def print_highscores(hs,s,u):
    hsp = popuplib.easymenu('highscores_popup','_popup_choice',None)
    hsp.settitle('Hangman Highscores')
    for rank in hs:
        if rank['steamid'] == s:
            hsp.addoption('',number(rank) + ': ' + rank['name'] + ' ' + str(rank['score']) + ' points ( ' + str(rank['won']) + '/' + str(rank['lost']) + ') <--')
        else:
            hsp.addoption('',number(rank) + ': ' + rank['name'] + ' ' + str(rank['score']) + ' points ( ' + str(rank['won']) + '/' + str(rank['lost']) + ')')
    hsp.send(u)

### Fiveline Hangman Grafic Engine (lol @ engine) ###
def fivelines(u,e):
    for line in gfx[5][e]:
        es.tell(u,line)

### Eightline Hangman Grafic Engine (lol @ engine) ###
def eightlines(u,e):
    global hmp
    hmp = popuplib.create('hangman_popup')
    for line in gfx[8][e]:
        hmp.addline(line)
    hmp.send(u)    

### Randomly get a word ###
def get_word(s):
    if len(user[s]['words']) != 0:
        word = user[s]['words'][0]
        while word in user[s]['words']:
            word = words[randint(0,len(words) - 1)]
    else:
        word = words[randint(0,len(words) - 1)]
    return word

### Get the rank of a user ###
def get_rank(s):
    temp = []
    for x in user:
        score = user[x]['won'] / user[x]['lost']
        temp.append([score,x])
    temp.sort()
    for x in range(len(temp)):
        if temp[x][1] == s:
            return x + 1
            
### Get the highscores ###
def get_highscores():
    temp = []
    for x in user:
        score = user[x]['won'] / user[x]['lost']
        temp.append([score,user[x]['won'],user[x]['lost'],user[x]['name'],user[x]])
    temp.sort()
    highscores = {}
    for x in len(temp):
        highscores[x] = {'score': temp[0],'won':temp[1],'lost':temp[2],'name':temp[3],'steamid':temp[4]}
    return highscors    

### Load the GFX from a file ###
def load_gfx():
    global gfx
    gfx = {5:{1:'1',2:'2',3:'3',4:'4',5:'5'},8:{1:'1',2:'2',3:'3',4:'4',5:'5',6:'6',7:'7',8:'8'}}
    if lines == 8:
        f = open(es.getAddonPath('hangman') + '/hangman_8.gfx','r')
        for x in range(1,9):
            gfx[8][x] = []
            for y in range(8):
                gfx[8][x].append(f.readline().replace('\n',''))
        f.close()
    else:
        f = open(es.getAddonPath('hangman') + '/hangman_5.gfx','r')
        for x in range(1,6):
            gfx[5][x] = []
            for y in range(5):
                gfx[5][x].append(f.readline().replace('\n',''))
        f.close()

### Load the user database ###
def load_db():
    global user
    if lexists(es.getAddonPath('hangman') + '/hangman_users.db'):
        f = open(es.getAddonPath('hangman') + '/hangman_users.db','rb')
        user = cPickle.load(f)
        f.close()
    else:
        user = {}

### Load pending games ###
def load_pending():
    global pending
    if lexists(es.getAddonPath('hangman') + '/hangman_pending.db'):
        f = open(es.getAddonPath('hangman') + '/hangman_pending.db','rb')
        pending = cPickle.load(f)
        f.close()
    else:
        pending = {}

### Load the configs ###
def load_configs():
    global config
    f = open(es.getAddonPath('').replace('\\','/').replace('addons/eventscripts/','cfg/hangman.cfg'),'r')
    lines = f.read().splitlines()
    config = {'adminpw':None,'wordlist':None}
    for line in lines:
        if line.startswith('adminpw'):
            config['adminpw'] = line[8:]
        elif line.startswith('wordlist'):
            config['wordlist'] = line[9:]
    if config['wordlist'] == None:
        if lexists(es.getAddonPath('hangman') + '/hangman_words.txt'):
            config['wordlist'] = 'hangman_words.txt'
        else:
            es.dbgmsg(1,'Shutting down hangman because of invalid or non existing wordlist')
            es.unload('hangman')
    if not lexists(es.getAddonPath('hangman') + '/' + config['wordlist']):
        if lexists(es.getAddonPath('hangman') + '/hangman_words.txt'):
            config['wordlist'] = 'hangman_words.txt'
        else:
            es.dbgmsg(1,'Shutting down hangman because of invalid or non existing wordlist')
            es.unload('hangman')

### load the wordlist ###
def load_words():
    global words, listversion
    words = []
    f = open(es.getAddonPath('hangman') + '/' + config['wordlist'],'r')
    for line in f.readlines():
        words.append(line.replace('\n',''))

### Save the user database ###
def save_db():
    f = open(es.getAddonPath('hangman') + '/hangman_users.db','wb')
    cPickle.dump(user,f)
    f.close()

### Save the pending games ###
def save_pending():
    f = open(es.getAddonPath('hangman') + '/hangman_pending.db','wb')
    cPickle.dump(pending,f)
    f.close()    
    
### Nicely format the numbers ###
def number(nr):
    nr = str(nr)
    if nr.endswith('1'):
        return nr + 'st'
    elif nr.endswith('2'):
        return nr + 'nd'
    elif nr.endswith('3'):
        return nr + 'rd'
    else:
        return nr + 'th'